use std::path::PathBuf;
use std::sync::atomic::AtomicBool;
use std::sync::Arc;
use std::thread;

use color_eyre::eyre::Context;
use color_eyre::Result;
use crossbeam::channel::bounded;
use crossbeam::select;
use signal_hook::consts::TERM_SIGNALS;
use signal_hook::flag;
use signal_hook::iterator::Signals;

use crate::config::Config;
use crate::daemon::worker::TaskScheduler;

use self::listener::Listener;

mod listener;
pub mod worker;

#[tracing::instrument]
pub fn handle(runtime_dir: PathBuf, config_path: Option<PathBuf>, session: String) -> Result<()> {
    // Ensure that sending a termination signal twice will immediately kill
    {
        let term_now = Arc::new(AtomicBool::new(false));
        for sig in TERM_SIGNALS {
            flag::register_conditional_shutdown(*sig, 1, Arc::clone(&term_now))?;
            flag::register(*sig, Arc::clone(&term_now))?;
        }
    }

    let config = Config::load(config_path).wrap_err("Failed to load config")?;
    tracing::debug!(?config);

    let rx_signals = {
        let mut signals = Signals::new(TERM_SIGNALS)?;
        let (tx, rx) = bounded(1);

        thread::Builder::new()
            .name("signal-handler".into())
            .spawn(move || {
                for sig in &mut signals {
                    tx.send(sig).expect("Failed to send signal");
                }
            })
            .wrap_err("Failed to start signal handler")?;

        tracing::trace!("Started signal handler");

        rx
    };

    let mut scheduler = TaskScheduler::new(config.workers, config.tokens)
        .wrap_err("Failed to create task scheduler")?;
    let listener = Listener::new(runtime_dir, session).wrap_err("Failed to create listener")?;

    loop {
        select! {
            recv(listener.receiver()) -> task => scheduler.schedule(task?),
            recv(rx_signals) -> _ => {
                tracing::info!(
                    "Received shutdown signal, waiting for workers to finish. \
                        Send termination again to force quit.");
                break;
            },
        }
    }

    scheduler.terminate();

    Ok(())
}
